Package: Rbec
Type: Package
Title: Rbec: a tool for analysis of amplicon sequencing data from
        synthetic microbial communities
Version: 1.20.0
Authors@R: person("Pengfan", "Zhang", email = "pzhang@mpipz.mpg.de", role = c("aut", "cre"))
Description: Rbec is a adapted version of DADA2 for analyzing amplicon sequencing data
    from synthetic communities (SynComs), where the reference sequences for each strain exists.
    Rbec can not only accurately profile the microbial compositions in SynComs, but also predict 
    the contaminants in SynCom samples.
License: LGPL-3
Imports: Rcpp (>= 1.0.6), dada2, ggplot2, readr, doParallel, foreach,
        grDevices, stats, utils
LinkingTo: Rcpp
RoxygenNote: 7.1.1
biocViews: Sequencing, MicrobialStrain, Microbiome
Suggests: knitr, rmarkdown
VignetteBuilder: knitr
git_url: https://git.bioconductor.org/packages/Rbec
git_branch: RELEASE_3_23
git_last_commit: 1d7e45a
git_last_commit_date: 2026-04-28
Repository: Bioconductor 3.23
Date/Publication: 2026-04-28
NeedsCompilation: yes
Packaged: 2026-04-29 00:06:25 UTC; biocbuild
Author: Pengfan Zhang [aut, cre]
Maintainer: Pengfan Zhang <pzhang@mpipz.mpg.de>

useDynLib(Rbec, .registration=TRUE)
importFrom(Rcpp, evalCpp)
importFrom("grDevices", "dev.off", "pdf")
importFrom("stats", "IQR", "dpois", "loess", "predict")
importFrom("utils", "write.table")
import(dada2)
import(readr)
import(ggplot2)
import(doParallel)
import(foreach)
export(Rbec)
export(Contam_detect)

# Rbec: a tool for analysis of amplicon sequencing data from synthetic microbial communities

Our manuscript is currently available as a preprint in [*bioRxiv*](https://doi.org/10.1101/2021.01.15.426834).

Rbec is a tool for analysing amplicon sequencing data from synthetic communities (SynComs), where the reference sequences for each strain are already available. Rbec can accurately correct PCR and sequencing errors, identify intra-species polymorphic variation, and detect contaminations in SynCom amplicon data.

Installation
---

Rbec is a free R package. To install Rbec, one needs to have the 'dada2' package (https://benjjneb.github.io/dada2/dada-installation.html) installed beforehand manually with the following command:
```
if (!requireNamespace("BiocManager", quietly = TRUE))
    install.packages("BiocManager")
BiocManager::install("dada2")
```

Then you can directly download with devtools:
```
devtools::install_github("PengfanZhang/Rbec", dependencies = TRUE)
```

Alternatively, you can download from Bioconductor:
```
if (!requireNamespace("BiocManager", quietly = TRUE))
    install.packages("BiocManager")
BiocManager::install("Rbec")
```

How to use Rbec?
---

1. Characterizing microbial communities in SynComs
---

Rbec can use as an input FASTQ files from either single- or pair-end sequencing data. In addition, a database of reference sequences in FASTA format needs to be provided.

Amplicon reads should be manually filtered by excluding reads with ambiguous reads with USEARCH or a similar software before running Rbec. Following is an example of removing reads with ambiguous reads with USEARCH:
```
usearch -fastq_filter <input_fastq> -fastqout <output_fastq> -fastq_maxns 0
```

Sequences in the reference database should be already truncated to the region matching the amplicon reads, which should be flanked by the primer sequences. For example, If we sequence the V5-V7 region of the synthetic bacterial community, the reference sequence of each strain in the reference database should also be truncated to retain the V5-V7 region only rather than the full-length of *16S* rRNA sequences. To truncate the reference sequences into a specific region, a tool such as cutadapt can be used:

```
cutadapt -g <forward_sequencing_primer> -a <reverse_sequencing_primer> -n 2 -o <output> <input>
```

Rbec currently only supports the reference database in a non-wrapped format. Towards this end, the following Unix command can be used to transform the wrapped sequences into non-wrapped ones.

```
awk 'BEGIN{seqs=""}{if(/^>/){if(seqs!=""){print seqs;seqs=""}; print $0} else{seqs=seqs""$0}}END{print seqs}' <input_reference_database> > <output_reference_database>
```

After preparation of the input files, Rbec can be run to estimate SynCom community profiles. An example with a small test dataset from a single bacterial strain illustrates this process:

```
fq <- system.file("extdata", "test_raw_merged_reads.fastq.gz", package="Rbec")

ref <- system.file("extdata", "test_ref.fasta", package="Rbec")

Rbec(fq, ref, tempdir(), 1, 500, 33)
```

Users can merge the profiling tables from different samples into a single one by using 'mergeSequenceTables' function in DADA2.

2. Detecting contaminants
---

One of the main sources of technical variation in gnotobiotic experiments is caused by microbial contaminations occurring during the development of the experiment or already present during input SynCom preparation. One of the features of Rbec is the assessment of likely contaminated samples based the recruitment ratio of sequencing reads across samples. When analysing data with Rbec, a separate log file is provided as an output for each sample. To predict contaminated samples, a text file containing a list of all paths to the log files needs to be provided before running the following command:

```
log_path <- list.files(paste(path.package("Rbec"), "inst/extdata/contamination_test", sep="/"), recursive=TRUE, full.names=TRUE)

log_file <- tempfile()

writeLines(log_path, log_file)

Contam_detect(log_file, tempdir())
```

This command will generate a plot showing the distribution of percentages of corrected reads across the whole sample set and a log file with predicted contaminated samples are generated. As a general rule, 90% or more of reads should be corrected in clean SynCom samples.

3. Rbec dependence on accurate reference sequences
---

Since Rbec is a reference-based method for error correction in sequencing reads, the accuracy of the reference sequence would critically influence the result. Inference of marker gene sequences from draft genome assemblies or via Sanger sequencing might lead to errors that may negatively impact the accuracy of the results. If errors are present in the reference sequence of a certain strain and no reads can be perfectly aligned to that reference (with an initial abundance of 0 for that reference), Rbec flags this strain as absent with an abundance of 0. One tricky way to overcome this issue in which references are not 100% accurate is that you can look up the contamination sequences outputted by the Rbec function at the first round and align the contamination sequences to the references of strains that are missing in the community. When this occurs, the correct sequence will be flagged by Rbec as a putative contaminant. The user can use this information to replace the erroneous entry in the reference database and re-run Rbec. This should be done with care to avoid true contaminants to be mistaken by members of the original input SynCom when their phylogenetic distance is low.

Credits
---

Pengfan Zhang (pzhang@mpipz.mpg.de); Ruben Garrido-Oter (garridoo@mpipz.mpg.de)

#' Reference-based error correction of amplicon sequencing data
#'
#' @description
#' This function calculates the abundance probabilities for each reads using poisson distribution
#'
#' @details Ruben Garrido-Oter's group, Plant-Microbe interaction, Max Planck Institute for Plant Breeding Research
#' @author  Pengfan Zhang
#'
#' @param derep dereplicated reads (Ns are not allowed in the reads)
#' @param ref the unique reference sequences of the reference seqeunces, each sequence must be in one line (Ns are not allowed in the sequences)
#' @param error_matrix The error matrix from the former iteration
#'
#' @usage abd_prob(derep, ref, error_matrix)
#'
#' @return Returns the lambda value and pvalue for each reads
#'
#' @noRd
#'

abd_prob <- function(derep, ref, error_matrix) {

    lambda_out <- data.frame(numeric(0))
    seqs <- names(derep[["uniques"]])
    abd <- derep[["uniques"]]
    seqs <- toupper(seqs)
    bestref <- derep[["bestref"]]

    input_data <- data.frame(seqs, abd, bestref, stringsAsFactors=FALSE)

    # calculate the estimated abundance of reference sequences by summing up the abundances of
    # reads potentailly produced from that reference sequence
    ref_abd <- data.frame(bestref, abd)
    ref_abd <- data.frame(tapply(ref_abd[, -1], INDEX=list(ref_abd[,1]), sum))
    ref$est_abd[match(rownames(ref_abd), ref[, 1])] <- ref_abd[, 1]

    # function for lambda calculation
    lambda_calculation <- function(seq, abd, bestref) {
        best_match <- match(bestref, ref[, 1])
        align_out <- nwalign(seq, bestref)
        Seq1align <- align_out[1]
        Seq2align <- align_out[2]
        Seq1align <- unlist(strsplit(Seq1align, ""))
        Seq2align <- unlist(strsplit(Seq2align, ""))
        align_pairs <- paste(Seq2align, Seq1align, sep="2")
        align_pairs <- align_pairs[! endsWith(align_pairs, "-")]
        qual <- round(derep[["quals"]][match(seq, names(derep[["uniques"]])), ], digits=0)
        qual <- as.character(qual[!is.na(qual)])
        retain_index <- which(!align_pairs %in% c("A2A", "C2C", "G2G", "T2T"))
        align_pairs <- align_pairs[retain_index]
        qual <- qual[retain_index]
        data4lambda <- data.frame(align_pairs, qual)
        tp <- apply(data4lambda, 1, function(x) error_matrix[x[1], x[2]])
        lambda0 <- prod(tp)
        lambda_out <- rbind(c(lambda0, best_match, abd))
        return(lambda_out)
    }

    # calculate lambdas for each dereplicated sequence
    lambda_out <- apply(input_data, 1, function(x) lambda_calculation(x[1], x[2], x[3]))
    lambda_out <- t(lambda_out)
    lambda_out <- apply(lambda_out, 2, function(x) as.numeric(x))
    lambda_out <- as.data.frame(lambda_out, stringsAsFactors=FALSE)

    # calculate abundance probabilities
    names(lambda_out) <- c("Lambda", "Ref_ID", "Obs_abd")
    lambda_out$ref_abd <- ref[lambda_out[, 2], "est_abd"]
    lambda_out$E <- lambda_out$Lambda*lambda_out$ref_abd
    lambda_out$pval <- apply(lambda_out, 1, function(x) 1-sum(dpois(0:(x[3]-1), x[5])))
    return(lambda_out)
}

#' Reference-based error correction of amplicon sequencing data
#'
#' @description
#' This function count the transition matrix
#'
#' @details Ruben Garrido-Oter's group, Plant-Microbe interaction, Max Planck Institute for Plant Breeding Research
#' @author Pengfan Zhang
#'
#' @param query list containing subsampled amplicon sequencing reads, quality scores, and reference sequences showing the highest identity for each read (Ns are not allowed in the reads)
#' @param ascii ascii characters used to encode phred scores
#'
#' @usage trans_m(query, ascii)
#'
#' @return The output is a 20 by 43 matrix containing the counts for different kinds of transitions
#'
#' @noRd
#'

trans_m <- function(query, ascii) {

    # initialize the transition matrix
    trans_matrix <- matrix(0, ncol=43, nrow=20)
    trans_kinds <- c(paste(rep(c("A","T","G","C","-"), each=4),
                           rep(2, 20), rep(c("A","T","G","C"), 5), sep=""))
    rownames(trans_matrix) <- trans_kinds
    colnames(trans_matrix) <- as.character(seq(0, 42))

    tm_cal <- function(que, qual, ref){

        # perform global alignment between query and reference using NW
        align_out <- nwalign(que, ref)
        Seq1align <- align_out[1]
        Seq2align <- align_out[2]
        Seq1align <- unlist(strsplit(Seq1align, ""))
        Seq2align <- unlist(strsplit(Seq2align, ""))
        align_pairs <- paste(Seq2align, Seq1align, sep="2")
        align_pairs <- align_pairs[!endsWith(align_pairs, "-")]

        # check if the quality system is phred 33 or 64
        qual <- as.character(strtoi(charToRaw(paste((qual), collapse="")), 16L)-ascii)

        align_data <- data.frame(alig=align_pairs, qual=qual, stringsAsFactors=FALSE)
        return(align_data)
    }

    # for all query sequences, calculate global alignments to reference sequence and corresponding transitions
    align_data <- do.call(rbind, apply(query, 1, function(x) tm_cal(x[1], x[2], x[3])))

    # transform transitions for each aligned querity to transition matrix
    align_data <- data.frame(table(align_data), stringsAsFactors=FALSE)
    for(i in seq_len(nrow(align_data))){
        trans_matrix[as.character(align_data[i, 1]), as.character(align_data[i, 2])] <- align_data[i, 3]
    }

    return(trans_matrix)

}

#' Reference-based error correction of amplicon sequencing data
#'
#' @description
#' This function corrects the amplicon sequencing data from synthetic communities where the reference sequences are known a priori
#'
#' @details Ruben Garrido-Oter's group, Plant-Microbe interaction, Max Planck Institute for Plant Breeding Research
#' @author  Pengfan Zhang
#'
#' @param fastq the path of the fastq file containg merged amplicon sequencing reads (Ns are not allowed in the reads)
#' @param reference the path of the unique reference sequences, each sequence must be in one line (Ns are not allowed in the sequences)
#' @param outdir the output directory, which should be created by the user
#' @param threads the number of threads used, default 1
#' @param sampling_size the sampling size for calculating the error matrix, default 5000
#' @param ascii ascii characters used to encode phred scores (33 or 64), default 33
#' @param min_cont_obs_abd the minimum oberseved abundace of unique tags for detecting contamination sequences, default 200
#' @param min_cont_abd the relative abundance of unique tgas for detecting contamination sequences that can't be corrected by any of the references, default 0.03
#' @param min_E the minimum expectation of the Possion distribution for the identification of paralogues, default 0.05
#' @param min_P the minimum P value threshold of the Possion distribution to correct a read, default 1e-40
#' @param ref_seeker the method for finding the candidate error-producing reference sequence for a tag showing identical lowest K-mer distance to multiple references. 1 for the abundance-based method; 2 for the transition probability-based method, default 1.
#' @param cn the copy number table documenting the copy number of the marker gene in each strain. Rbec will normalize the strain abundance if the copy number is available
#'
#' @import readr
#' @import dada2
#' @import doParallel
#' @import foreach
#'
#' @usage Rbec(fastq, reference, outdir, threads=1, sampling_size=5000, ascii=33, min_cont_obs_abd=200, min_cont_abd=0.03, min_E=0.05, min_P=1e-40, ref_seeker=1, cn=NULL)
#' @examples
#' fastq <- system.file("extdata", "test_raw_merged_reads.fastq.gz", package = "Rbec")
#'
#' ref <- system.file("extdata", "test_ref.fasta", package = "Rbec")
#'
#' Rbec(fastq=fastq, reference=ref, outdir=tempdir(), threads=1, sampling_size=500, ascii=33)
#'
#' @return lambda_final.out the lambda value and pvalue of the Poisson distribution for each read
#' @return error_matrix_final.out the error matrix in the final iteration
#' @return strain_table.txt the strain composition of the sample
#' @return strain_table_normalized.txt the copy-number-normalized strain composition of the sample if the copy number table is provided
#' @return contamination_seq.fna the potential sequences generated by contaminants
#' @return rbec.log percentage of corrected reads, which can be used to predict contaminated samples
#' @return paralogue_seq.fna paralogue sequences found in each strain except for the reference provided 
#'
#' @export
#'

Rbec <- function(fastq, reference, outdir, threads=1, sampling_size=5000, ascii=33, min_cont_obs_abd=200, min_cont_abd=0.03, min_E=0.05, min_P=1e-40, ref_seeker=1, cn=NULL){


    t1 <- Sys.time()

    # parse input parameters and read reference sequences

    tryCatch(
        para_check <- as.numeric(threads),
        error = function(e) {stop("parameters 'threads' should be intergers!")},
        warning =  function(w) {stop("parameters 'threads' should be intergers!")}
    )

    tryCatch(
        para_check <- as.numeric(sampling_size),
        error = function(e) {stop("parameters 'sampling_size' should be intergers!")},
        warning =  function(w) {stop("parameters 'sampling_size' should be intergers!")}
    )

    tryCatch(
        para_check <- as.numeric(ascii),
        error = function(e) {stop("parameters 'ascii' should be intergers!")},
        warning =  function(w) {stop("parameters 'ascii' should be intergers!")}
    )


    ref <- read_lines(reference)

    ref_name <- data.frame(name=ref[which(seq_along(ref)%%2==1)], seq=toupper(ref[which(seq_along(ref)%%2==0)]), stringsAsFactors=FALSE)
    ref_name$name <- sub(">", "", ref_name$name)
    ref <- data.frame(ref_seq=toupper(ref[which(seq_along(ref)%%2==0)]), stringsAsFactors=FALSE)

    # calculate the error matrix
    error_ref_matrix <- error_m(fastq, ref, sampling_size, threads, ascii, ref_seeker)
    error_matrix <- error_ref_matrix[["err"]]
    ref <- error_ref_matrix[["ref"]]
    derep <- error_ref_matrix[["derep"]]
    message("Finished calculating the transition and error matrices.")

    # calculate abundance probabilities
    lambda_out <- abd_prob(derep, ref, error_matrix)

    final_list <- consis_err(fastq, derep, ref, lambda_out, sampling_size, ascii, min_E, min_P)

    straintab <- final_list[["table"]]
    straintab[, 1] <- ref_name$name[match(straintab[, 1], ref_name[, 2])]
    # normalize the abundance if copy number is available
    if (!is.null(cn)){
      cp_n <- as.data.frame(read_delim(cn, delim="\t"), stringsAsFactors = FALSE)
      straintab_norm <- straintab
      straintab_norm$Corrected_Abundance <- round(straintab_norm$Corrected_Abundance/cp_n[match(straintab_norm$Ref_seq, cp_n[, 1]), 2])
    }

    lambda_out <- final_list[["lambda_out"]]
    lambda_out[, 2] <- ref_name$name[match(lambda_out[, 2], ref_name[, 2])]

    # identify potential contaminants from reads which cannot be corrected and have high abundance (larger than min_cont_abd)
    contamination <- which(final_list$lambda_out$Corrected=="No" &
                         (final_list$lambda_out$Obs_abd >= min_cont_obs_abd | final_list$lambda_out$Obs_abd/error_ref_matrix$total_reads>=min_cont_abd))
    # output sequences of potential contaminants
    conta_out <- data.frame(stringsAsFactors=FALSE)
    c <- 1
    for (i in contamination){
	conta_align <- nwalign(names(derep$uniques[i]), final_list$lambda_out$Ref_ID[i])
	conta_align1 <- unlist(strsplit(conta_align[1], ""))
	conta_align2 <- unlist(strsplit(conta_align[2], ""))
	iden <- paste(round(length(which(conta_align1==conta_align2))/length(conta_align1)*100, 1), "%", sep="")
	conta_iden <- paste(lambda_out$Ref_ID[i], iden, sep=":")
        conta_rec <- paste(">Contamination_seq", c, sep="")
	conta_abd <- paste("Observed Abundace:", derep$uniques[i], sep="")
        conta_rate <- paste("Relative Abundance:", round(derep$uniques[i]/error_ref_matrix$total_reads*100, 1), "%", sep="")
        conta_rec <- paste(conta_rec, conta_abd, conta_rate, conta_iden, sep="; ")
        conta_out <- rbind(conta_out, conta_rec, stringsAsFactors = FALSE)
        conta_rec <- names(derep$uniques[i])
        conta_out <- rbind(conta_out, conta_rec, stringsAsFactors = FALSE)
        c <- c+1
    }
    
    #output paralogue sequences for each strain
    paralog_out <- data.frame(stringsAsFactors=FALSE)
    c <- 1
    for (i in seq(nrow(lambda_out))){
      if (lambda_out$Lambda[i]!=1 & lambda_out$E[i]>=min_E & lambda_out$Corrected[i]=="Yes" & lambda_out$Obs_abd[i]/max(lambda_out$Obs_abd[which(lambda_out$Ref_ID==lambda_out$Ref_ID[i])])>=1/15){
        paralog_rec <- paste(paste(">", "Seq_", c, sep=""), paste("Paralogue", lambda_out$Ref_ID[i], sep = "="), sep = ";")
        paralog_out <- rbind(paralog_out, paralog_rec, stringsAsFactors=FALSE)
        paralog_out <- rbind(paralog_out, names(derep$uniques[i]), stringsAsFactors=FALSE)
        c <- c+1
      }
    }

    # output final tables and logs

    dir.create(outdir)
    lambdaout <- paste(outdir, "lambda_final.out", sep="/")
    emout <- paste(outdir, "error_matrix_final.out", sep="/")
    straintabout <- paste(outdir, "strain_table.txt", sep="/")
    straintabout_norm <- paste(outdir, "strain_table_normalized.txt", sep="/")
    contaout <- paste(outdir, "contamination_seq.fna", sep="/")
    paralogout <- paste(outdir, "paralogue_seq.fna", sep="/")
    logout <- paste(outdir, "rbec.log", sep="/")

    write.table(lambda_out, file= lambdaout, quote=FALSE, sep="\t", row.names = FALSE)
    write.table(final_list[["err"]], file= emout, quote=FALSE, sep="\t")
    write.table(straintab, file= straintabout, quote=FALSE, sep="\t", row.names=FALSE)
    if (!is.null(cn)){
      write.table(straintab_norm, file= straintabout_norm, quote=FALSE, sep="\t", row.names=FALSE)
    }
    write.table(conta_out, file = contaout, quote=FALSE, sep="", col.names = FALSE, row.names = FALSE)
    write.table(paralog_out, file=paralogout, quote=FALSE, sep="", col.names = FALSE, row.names = FALSE)

    t2 <- Sys.time()
    time.used <- difftime(t2, t1, units = "mins")
    time.used <- sub("Time difference of ","",time.used)
    time.used <- paste("Total time used:", time.used, "mins", sep=" ")
    summ <- paste(final_list$total_corrected/error_ref_matrix$total_reads*100, "% of reads were corrected!", sep="")

    log_out <- rbind(time.used, summ)
    write.table(log_out, file = logout, quote=FALSE, sep="", col.names = FALSE, row.names = FALSE)

}

## ----include = FALSE----------------------------------------------------------
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>"
)

## ----CommunityProfiling, eval = TRUE------------------------------------------
library(Rbec)
fq <- system.file("extdata", "test_raw_merged_reads.fastq.gz", package="Rbec")

ref <- system.file("extdata", "test_ref.fasta", package="Rbec")

Rbec(fq, ref, tempdir(), 1, 500, 33)

## ----Contam_detect, eval = TRUE-----------------------------------------------
log_path <- list.files(paste(path.package("Rbec"), "extdata/contamination_test", sep="/"), recursive=TRUE, full.names=TRUE)
log_file <- tempfile()
writeLines(log_path, log_file)
Contam_detect(log_file, tempdir())

#' Reference-based error correction of amplicon sequencing data
#'
#' @description
#' This function iterates the error matrix till reaching the stable stage
#'
#' @details Ruben Garrido-Oter's group, Plant-Microbe interaction, Max Planck Institute for Plant Breeding Research
#' @author  Pengfan Zhang
#'
#' @param fq the path of the fastq file (Ns are not allowed in the reads)
#' @param derep the derepliacted reads by dada2 function
#' @param ref the reference sequences, each sequence should take up one line (Ns are not allowed in the reads)
#' @param lambda_out the matrix containg lambda value and pvalue from the former iteration
#' @param sampling_size the subsampling size of the reads
#' @param ascii ascii characters used to encode phred scores
#' @param min_E the minimum expectation value of the Possion distribution for detecting paralogs within the same strain
#' @param  min_P the P value cutoff for identifying errouneous reads
#' @param max_diff_abs the maximum absolute difference in number of corrected reads between two iterations, together with max_diff_ratio, before jumping out of the iiteration
#' @param max_diff_ratio the maximum difference in the percentages of corrected reads between two iterations
#'
#' @usage consis_err(fq, derep, ref, lambda_out, sampling_size, ascii, min_E=0.05, min_P=1e-40, max_diff_abs=100, max_diff_ratio=0.01)
#'
#' @return Returns the final files
#'
#' @noRd
#'

consis_err <- function(fq, derep, ref, lambda_out, sampling_size, ascii, min_E=0.05, min_P=1e-40, max_diff_abs=100, max_diff_ratio=0.01){
    raw_data <- read_lines(gzfile(fq))
    n <- 1
    repeat {
        message("Start ", n ," iterations")

        # calculate total number of corrected reads and total abundances from the previous iteration
        previous_round <- which(lambda_out$E>=min_E | lambda_out$pval>=min_P)
        previous_sum <- sum(lambda_out[previous_round, "Obs_abd"])

        # update abundances of reference sequences from erroneous reads asigned to them
        est_abd <- numeric(0)
        for (i in seq_along(unique(lambda_out$Ref_ID))) {

            # find all sequences with high expectation and P-val which could have been generated by a given reference sequence
            idx <-which((lambda_out$E>=min_E | lambda_out$pval>=min_P) & lambda_out$Ref_ID==unique(lambda_out$Ref_ID)[i])

            # calculate estimated abundances of a given reference sequence by adding up all the abundances of assigned erroneous sequences
            ref_i_abd <- sum(lambda_out$Obs_abd[idx])
            est_abd <- c(est_abd, ref_i_abd)

        }

        # update estimated abundances for each reference sequence
        ref$est_abd[unique(lambda_out$Ref_ID)] <- est_abd
        message(ref$est_abd)

        # extract raw sequences that belong to the error-corrected unique sequences found in the previous round
        current_round <- which(derep[["map"]] %in% previous_round)

        # sample the extracted sequences for transition matrix generation
        current_sample <- sample(current_round, sampling_size)
        seqs <- raw_data[current_sample*4-2]
        seqs <- toupper(seqs)
        quals <- raw_data[current_sample*4]
        bestref <- derep[["bestref"]][derep[["map"]][current_sample]]

        # calculate transition matrix and update error matrix
        query <- data.frame(seqs=seqs, qual=quals, ref=bestref)
        trans_matrix <- trans_m(query, ascii)
        error_matrix <- loessErr(trans_matrix)
        lambda_out <- abd_prob(derep, ref, error_matrix)

        # calculate total number of corrected reads and total abundances from the current iteration
        current_round <- which(lambda_out$E>=min_E | lambda_out$pval>=min_P)
        current_sum <- sum(lambda_out[current_round, "Obs_abd"])

        # check for convergence
        diff_abs <- abs(previous_sum - current_sum)
        diff_ratio <- diff_abs/previous_sum

        if (diff_abs < max_diff_abs & diff_ratio <= max_diff_ratio) {

            message(diff_abs)
            message(diff_ratio)
            message("The model reached consistency after ", n ," iteration(s)")

            final_abd <- numeric(0)
            for (i in seq_along(unique(lambda_out$Ref_ID))){

                # find all sequences with high expectation and P-val which could have been generated by a given reference sequence
                idx <- which((lambda_out$E>=min_E | lambda_out$pval>=min_P) & lambda_out$Ref_ID==unique(lambda_out$Ref_ID)[i])

                # calculate final abundances of a given reference sequence by adding up all the abundances of assigned erroneous sequences
                ref_i_abd <- sum(lambda_out$Obs_abd[idx])
                final_abd <- c(final_abd, ref_i_abd)

            }  

            out_table <- data.frame(Ref_seq=ref[unique(lambda_out$Ref_ID), 1], Corrected_Abundance=final_abd)
            lambda_out$Total_corrected <- sum(lambda_out[which(lambda_out$E>=min_E | lambda_out$pval>=min_P), "Obs_abd"])
            lambda_out$Corrected <- "No"
            lambda_out$Corrected[which(lambda_out$E>=min_E | lambda_out$pval>=min_P)] <- "Yes"
            lambda_out$Ref_ID <- ref[lambda_out$Ref_ID, 1]
            out_file <- list(lambda_out=lambda_out, err=error_matrix, table=out_table, total_corrected=sum(final_abd))
            return(out_file)

          break
        }
        else {

            message(diff_abs)
            message(diff_ratio)
            n <- n+1

        }
    }
}

#' Reference-based error correction of amplicon sequencing data
#'
#' @description
#' This function fits the loess regression to the error matrix (Parts of this script have been adapted from code originally published with DADA2 (Callahan et al., 2016).)
#'
#' @details Ruben Garrido-Oter's group, Plant-Microbe interaction, Max Planck Institute for Plant Breeding Research
#' @author  Pengfan Zhang
#'
#' @param trans the transition matrix
#' @param min_err_rate the minimum transition probability for each substitution or insertion case
#'
#' @usage loessErr(trans, min_err_rate=1e-07)
#'
#' @return Returns the loess fitted error matrix
#'
#' @noRd
#'


loessErr <- function (trans, min_err_rate=1e-07) {
    qq <- as.numeric(colnames(trans))
    est <- matrix(0, nrow = 0, ncol = length(qq))
    for (nti in c("A", "C", "G", "T")) {
        for (ntj in c("A", "C", "G", "T")) {
            if (nti != ntj) {
                errs <- trans[paste0(nti, "2", ntj), ]
                tot <- colSums(trans[paste0(nti, "2", c("A", "C", "G", "T")), ])
                rlogp <- log10((errs + 1)/tot)
                rlogp[is.infinite(rlogp)] <- NA
                df <- data.frame(q = qq, errs = errs, tot = tot,
                         rlogp = rlogp)
                mod.lo <- loess(rlogp ~ q, df, weights = tot)
                pred <- predict(mod.lo, qq)
                maxrli <- max(which(!is.na(pred)))
                minrli <- min(which(!is.na(pred)))
                pred[seq_along(pred) > maxrli] <- pred[[maxrli]]
                pred[seq_along(pred) < minrli] <- pred[[minrli]]
                est <- rbind(est, 10^pred)
            }
        }
    }

    # add the insertion probabilities (-2A/(-2A+A2A+T2A+C2A+G2A))
    for (ntj in c("A", "C", "G", "T")) {
        errs <- trans[paste0("-", "2", ntj), ]
        tot <- colSums(trans[paste0(c("A", "C", "G", "T", "-"), "2", ntj), ])
        rlogp <- log10((errs + 1)/tot)
        rlogp[is.infinite(rlogp)] <- NA
        df <- data.frame(q = qq, errs = errs, tot = tot,
                     rlogp = rlogp)
        mod.lo <- loess(rlogp ~ q, df, weights = tot)
        pred <- predict(mod.lo, qq)
        maxrli <- max(which(!is.na(pred)))
        minrli <- min(which(!is.na(pred)))
        pred[seq_along(pred) > maxrli] <- pred[[maxrli]]
        pred[seq_along(pred) < minrli] <- pred[[minrli]]
        est <- rbind(est, 10^pred)
    }

    est[est < min_err_rate] <- min_err_rate

    # calculate the transition probabilities for ('A2A', 'T2T', 'C2C', 'G2G') and fill out the error matrix
    err <- rbind(1-colSums(est[1:3, ]), est[1:3, ], est[4, ], 1-colSums(est[4:6, ]), est[5:6, ], est[7:8, ],
               1-colSums(est[7:9, ]), est[9, ], est[10:12, ], 1-colSums(est[10:12, ]), est[13:16, ])

    rownames(err) <- paste0(rep(c("A", "C", "G", "T", "-"), each = 4),
                          "2", c("A", "C", "G", "T"))

    colnames(err) <- colnames(trans)
    return(err)

}

#' Reference-based error correction of amplicon sequencing data
#'
#' @description
#' This function calculate the error matrix
#'
#' @details Ruben Garrido-Oter's group, Plant-Microbe interaction, Max Planck Institute for Plant Breeding Research
#' @author  Pengfan Zhang
#'
#' @param fq the path of merged amplicon sequencing reads in fastq format (Ns are not allowed in the reads)
#' @param ref the unique reference sequences of the reference seqeunces, each sequence must be in one line (Ns are not allowed in the sequences)
#' @param sample_size the sampling size of reads to generate the transition matrix
#' @param threads the number of threads used to align the query reads to reference sequences
#' @param ascii ascii characters used to encode phred scores
#' @param ref_seeker the method for finding the candidate error-producing reference sequence for a tag showing identical lowest K-mer distance to multiple references. 1 for the abundance-based method; 2 for the transition probability-based method, 1 in default.
#'
#'
#' @import dada2
#' @import foreach
#' @import doParallel
#'
#' @usage error_m(fq, ref, sample_size, threads, ascii, ref_seeker)
#'
#' @return The output is a 20 by 43 transition probability matrix
#'
#' @noRd
#'

error_m <- function(fq, ref, sample_size=10000, threads, ascii, ref_seeker) {

    x <- NA
    t1 <- Sys.time()

    # read FASTQ file
    if (grepl(".gz", fq, fixed = TRUE)) {
        raw_data <- read_lines(gzfile(fq))
    }
    else {
        raw_data <- read_lines(fq)
    }
    if (sample_size > length(raw_data)/4){
      stop("The sampling size ", sample_size ," exceeds the total number of reads ", length(raw_data)/4 ," in the input file")
    }

    # calculate the initial abundance of reference sequences
    derep1 <- derepFastq(fq)
    ref$abd <- derep1[["uniques"]][match(ref$ref_seq, names(derep1[["uniques"]]))]
    ref <- ref[!is.na(ref$abd), ]
    ref$est_abd <- 0

    # find the reference sequences with highest similarities for each unique sequence
    uniseqs <- names(derep1[["uniques"]])

    findbest <- function(que) {
        # use the kmer distance function implemented in DADA2
        kdist <- function(que, ref, kmer=7) {
            dis <- kmer_dist(que, ref, kmer)
            return(dis)
        }
        # calculate k-mer distances for all sequences
        kdist_res <- unlist(lapply(ref$ref_seq, function(x) kdist(que, x)))
        best_match <- which.min(kdist_res)
        # if there are more than 1 hits showing the same maximum identity,
        # the reference with the highest abundance is chosen as the best reference sequence
        if (length(best_match) > 1) {
          if (ref_seeker==1){
		message("Searching best reference based on abundace information.")
            ref_candidate <- ref[best_match, ]
            ref_best <- ref_candidate[which.max(ref_candidate[, 2]), 1]
          }
          else{ # concatenate the reference sequences
		  message("Searching best reference based on transition probability.")
            ref_best <- paste(ref[best_match, 1], collapse = "/")
          }
        }
        else {
          ref_best <- ref[best_match, 1]
        }
        return(c(que, ref_best))
    }

    #cl <- makeCluster(threads)
    registerDoParallel(cores=threads)
    que2ref <- foreach::foreach(x=uniseqs, .combine='rbind') %dopar% findbest(x)
    stopImplicitCluster()
    que2ref <- as.character(que2ref[, 2])
    message("Finished finding the best reference sequences for each unique sequences.")
    t2 <- Sys.time()
    message(t2-t1)

    # sample raw sequences for transition matrix generation
    sample_label <- sample(seq_len(length(raw_data)/4), sample_size)
    seqs <- raw_data[sample_label*4-2]
    seqs <- toupper(seqs)
    qual <- raw_data[sample_label*4]
    bestref <- que2ref[derep1[["map"]][sample_label]]
    query <- data.frame(seqs=seqs, qual=qual, refs=bestref)
    if (ref_seeker==2 & length(which(grepl('/', query$refs)))>0 ){
      query <- query[-which(grepl('/', query$refs)), ]
    }

    # calculate the transition matrix and error matrix
    trans_matrix <- trans_m(query, ascii)
    error_matrix <- loessErr(trans_matrix)

    # choose the best ref with the highest transition probability for ref_seeker=2
    if (ref_seeker==2 & length(which(grepl('/', que2ref)))>0){
      for (i in which(grepl('/', que2ref))){
        lambda_choose <- c()
        ref_choose <- c()
        for (j in unlist(strsplit(que2ref[i], '/'))){
          align_out <- nwalign(derep1$uniques[i], j)
          Seq1align <- align_out[1]
          Seq2align <- align_out[2]
          Seq1align <- unlist(strsplit(Seq1align, ""))
          Seq2align <- unlist(strsplit(Seq2align, ""))
          align_pairs <- paste(Seq2align, Seq1align, sep="2")
          align_pairs <- align_pairs[! endsWith(align_pairs, "-")]
          qual <- round(derep1[["quals"]][match(seq, names(derep1[["uniques"]])), ], digits=0)
          qual <- as.character(qual[!is.na(qual)])
          retain_index <- which(!align_pairs %in% c("A2A", "C2C", "G2G", "T2T"))
          align_pairs <- align_pairs[retain_index]
          qual <- qual[retain_index]
          data4lambda <- data.frame(align_pairs, qual)
          tp <- apply(data4lambda, 1, function(x) error_matrix[x[1], x[2]])
          lambda0 <- prod(tp)
          ref_choose <- c(ref_choose, j)
          lambda_choose <- c(lambda_choose, lambda0)
        }
        que2ref[i] <- ref_choose[which.max(lambda_choose)]
      }
    }
    
    derep1[["bestref"]] <- que2ref

    error_ref_matrix <- list(ref=ref, err=error_matrix, derep=derep1, total_reads=length(raw_data)/4)
    return(error_ref_matrix)

}

#' Reference-based error correction of amplicon sequencing data
#'
#' @description
#' This function is designed for predicting the contaminated samples
#'
#' @details Ruben Garrido-Oter's group, Plant-Microbe interaction, Max Planck Institute for Plant Breeding Research
#' @author  Pengfan Zhang
#'
#' @param log_file the file contains a list of log files of each sample outputted with Rbec function
#' @param outdir output directory
#' @param outlier_constant the multiplier of variance to define the outlier
#'
#' @usage Contam_detect(log_file, outdir, outlier_constant=1.5)
#' @examples
#' #log_file <- system.file("extdata", "rbec_test.list", package = "Rbec")
#' log_path <- list.files(paste(path.package("Rbec"), 
#' 	"extdata/contamination_test", sep="/"), 
#' 	recursive=TRUE, full.names=TRUE)
#' log_file <- tempfile()
#' writeLines(log_path, log_file)
#' Contam_detect(log_file, tempdir())
#'
#' @import ggplot2
#' @import readr
#' 
#' @return Returns a plot showing the distribution of percentage of corrected reads across the whole sample set and a summary file recording which samples might be contaminated
#'
#' @export


Contam_detect <- function(log_file, outdir, outlier_constant=1.5){

    log.file <- read_lines(log_file)

    samples <- c()
    ratio <- c()
    outlier <- NA

    # read percentages of corrected reads for each sample
    for (i in seq(length(log.file))){
        corr_ratio <- read_lines(log.file[i], skip = 1)
        corr_ratio <- sub("% of reads were corrected!", "", corr_ratio)
        sample_name <- unlist(strsplit(log.file[i], "/"))
        sample_name <- sample_name[length(sample_name)-1]
        samples <- c(samples, sample_name)
        ratio <- c(ratio, corr_ratio)
    }

    ratio <- as.numeric(ratio)
    samples <- samples[order(ratio)]
    ratio <- sort(ratio)
    data4plot <- data.frame(sample = samples, ratio = ratio)
    # calculate the lower bound of percentage of corrected reads in clean samples
    #lower_bd <- mean(ratio) - outlier_constant*var(ratio)
    #IQR is more sensitive to small sample size in identifying contaminated samples
    lower_bd <- mean(ratio) - outlier_constant*IQR(ratio)

    data4plot$outlier <- "False"
    data4plot$outlier[which(data4plot$ratio < lower_bd)] <- "True"
    outlier_num <- length(which(data4plot$ratio < lower_bd))
    data4plot$sample <- factor(data4plot$sample, levels = as.character(data4plot$sample))

    # plot the distribution of percentages of corrected reads across the sample set
    p <- ggplot(data4plot, aes(sample, ratio, color = outlier)) + geom_point(size= 0.8) + geom_hline(yintercept = lower_bd, linetype = 3, size = 0.5, color = "grey") + geom_vline(xintercept = outlier_num+0.5, linetype = 3, size = 0.5, color = "grey") + theme(axis.text.x = element_text(angle=90, size=2, hjust = 1), axis.ticks.x = element_blank(), legend.position = "NA", panel.background = element_rect(colour = "black", fill="white"), panel.grid = element_blank()) + scale_color_manual(values = c("blue", "red")) + xlab("") + ylab("Percentage of corrected reads(%)")
    plot_out <- paste(outdir, "summary_plot.pdf", sep = "/")
    pdf(plot_out,width=10, height=6)
    print(p)
    dev.off()


    # output the statistic information of samples
    max_r <- round(max(ratio), digits = 2)
    min_r <- round(min(ratio), digits = 2)
    mean_r <- round(mean(ratio), digits = 2)
    output <- c(paste("The maximum percentage of corrected reads: ", max_r, sep=""), paste("The minimum percentage of corrected reads: ", min_r, sep=""), paste("Mean: ", mean_r, sep=""), " ", " ", "Potentially contaminated samples (less than mean-1.5*IQR):", paste(as.character(data4plot$sample[seq(outlier_num)]), round(data4plot$ratio[seq(outlier_num)], digits = 2), sep=' '))
    stat_out <- paste(outdir, "summary.log", sep = "/")
    write.table(as.data.frame(output), file=stat_out, sep="\t", quote=FALSE, col.names = FALSE, row.names = FALSE)
}

# Generated by using Rcpp::compileAttributes() -> do not edit by hand
# Generator token: 10BE3573-1514-4C36-9D1C-5A225CD40393

#' @name kmer_dist
#' @title DADA2
#' @description Calculate the kmer distance between two sequences. This function is originally developed by DADA2 in 2012 and we adapt it here.
#'
#' @param s1 A \code{character(1)} of DNA sequence 1.
#' @param s2 A \code{character(1)} of DNA sequence 2.
#' @param kmer_size Kmer size.
#' @return The kmer distance between two sequences
#' @noRd
#'
kmer_dist <- function(s1, s2, kmer_size) {
    .Call(`_Rbec_kmer_dist`, s1, s2, kmer_size)
}

# Register entry points for exported C++ functions
methods::setLoadAction(function(ns) {
    .Call('_Rbec_RcppExport_registerCCallable', PACKAGE = 'Rbec')
})

---
title: "Rbec: a tool for analysis of amplicon sequencing data from synthetic microbial communities"
author: "Pengfan Zhang"
abstract: >
  Rbec is a tool for analysing amplicon sequencing data from synthetic communities (SynComs), 
  where the reference sequences for each strain are already available. Rbec can accurately 
  correct PCR and sequencing errors, identify intra-species polymorphic variation, and detect 
  contaminations in SynCom amplicon data.
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{Rbec}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r, include = FALSE}
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>"
)
```

# Overview

Rbec, which is the abbreviation of reference-based error correction of amplicon sequencing data
from synthetic microbial communities, is the first documented software exclusively developed 
for analyzing the amplicon sequencing data from SynComs. Rbec can not only output accurate microbial 
composition for each sample, but also predict the potential contaminants in the artificial system.

How to use Rbec?
---

1. Characterizing microbial communities in SynComs
---

Rbec can be run to estimate SynCom community profiles. An example with a small test dataset from a single bacterial strain illustrates this process:

```{r CommunityProfiling, eval = TRUE}
library(Rbec)
fq <- system.file("extdata", "test_raw_merged_reads.fastq.gz", package="Rbec")

ref <- system.file("extdata", "test_ref.fasta", package="Rbec")

Rbec(fq, ref, tempdir(), 1, 500, 33)
```

2. Detecting contaminants
---

One of the main sources of technical variation in gnotobiotic experiments is caused by microbial contaminations occurring during the development of the experiment or already present during input SynCom preparation. One of the features of Rbec is the assessment of likely contaminated samples based the recruitment ratio of sequencing reads across samples. When analysing data with Rbec, a separate log file is provided as an output for each sample. To predict contaminated samples, a text file containing a list of all paths to the log files needs to be provided before running the following command:

```{r Contam_detect, eval = TRUE}
log_path <- list.files(paste(path.package("Rbec"), "extdata/contamination_test", sep="/"), recursive=TRUE, full.names=TRUE)
log_file <- tempfile()
writeLines(log_path, log_file)
Contam_detect(log_file, tempdir())
```

This command will generate a plot showing the distribution of percentages of corrected reads across the whole sample set and a log file with predicted contaminated samples are generated. As a general rule, 90% or more of reads should be corrected in clean SynCom samples.

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/Rbec.R
\name{Rbec}
\alias{Rbec}
\title{Reference-based error correction of amplicon sequencing data}
\usage{
Rbec(fastq, reference, outdir, threads=1, sampling_size=5000, ascii=33, min_cont_obs_abd=200, min_cont_abd=0.03, min_E=0.05, min_P=1e-40, ref_seeker=1, cn=NULL)
}
\arguments{
\item{fastq}{the path of the fastq file containg merged amplicon sequencing reads (Ns are not allowed in the reads)}

\item{reference}{the path of the unique reference sequences, each sequence must be in one line (Ns are not allowed in the sequences)}

\item{outdir}{the output directory, which should be created by the user}

\item{threads}{the number of threads used, default 1}

\item{sampling_size}{the sampling size for calculating the error matrix, default 5000}

\item{ascii}{ascii characters used to encode phred scores (33 or 64), default 33}

\item{min_cont_obs_abd}{the minimum oberseved abundace of unique tags for detecting contamination sequences, default 200}

\item{min_cont_abd}{the relative abundance of unique tgas for detecting contamination sequences that can't be corrected by any of the references, default 0.03}

\item{min_E}{the minimum expectation of the Possion distribution for the identification of paralogues, default 0.05}

\item{min_P}{the minimum P value threshold of the Possion distribution to correct a read, default 1e-40}

\item{ref_seeker}{the method for finding the candidate error-producing reference sequence for a tag showing identical lowest K-mer distance to multiple references. 1 for the abundance-based method; 2 for the transition probability-based method, default 1.}

\item{cn}{the copy number table documenting the copy number of the marker gene in each strain. Rbec will normalize the strain abundance if the copy number is available}
}
\value{
lambda_final.out the lambda value and pvalue of the Poisson distribution for each read

error_matrix_final.out the error matrix in the final iteration

strain_table.txt the strain composition of the sample

strain_table_normalized.txt the copy-number-normalized strain composition of the sample if the copy number table is provided

contamination_seq.fna the potential sequences generated by contaminants

rbec.log percentage of corrected reads, which can be used to predict contaminated samples

paralogue_seq.fna paralogue sequences found in each strain except for the reference provided
}
\description{
This function corrects the amplicon sequencing data from synthetic communities where the reference sequences are known a priori
}
\details{
Ruben Garrido-Oter's group, Plant-Microbe interaction, Max Planck Institute for Plant Breeding Research
}
\examples{
fastq <- system.file("extdata", "test_raw_merged_reads.fastq.gz", package = "Rbec")

ref <- system.file("extdata", "test_ref.fasta", package = "Rbec")

Rbec(fastq=fastq, reference=ref, outdir=tempdir(), threads=1, sampling_size=500, ascii=33)

}
\author{
Pengfan Zhang
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/Rbec_summarise.R
\name{Contam_detect}
\alias{Contam_detect}
\title{Reference-based error correction of amplicon sequencing data}
\usage{
Contam_detect(log_file, outdir, outlier_constant=1.5)
}
\arguments{
\item{log_file}{the file contains a list of log files of each sample outputted with Rbec function}

\item{outdir}{output directory}

\item{outlier_constant}{the multiplier of variance to define the outlier}
}
\value{
Returns a plot showing the distribution of percentage of corrected reads across the whole sample set and a summary file recording which samples might be contaminated
}
\description{
This function is designed for predicting the contaminated samples
}
\details{
Ruben Garrido-Oter's group, Plant-Microbe interaction, Max Planck Institute for Plant Breeding Research
}
\examples{
#log_file <- system.file("extdata", "rbec_test.list", package = "Rbec")
log_path <- list.files(paste(path.package("Rbec"), 
	"extdata/contamination_test", sep="/"), 
	recursive=TRUE, full.names=TRUE)
log_file <- tempfile()
writeLines(log_path, log_file)
Contam_detect(log_file, tempdir())

}
\author{
Pengfan Zhang
}
